"""Gives users direct access to classes, functions, and decorator."""
from circuitry.circuitry import \
    bit, constant, input, input_one, input_two, output, \
    bits, constants, inputs, outputs, \
    synthesize



"""
Test suite containing examples that demonstrate how the library can be used
to synthesize circuits from functions.

To view the source code of an example function, click its **[source]** link.
These examples (as well as additional background information) are discussed
in more detail in a `relevant report <https://eprint.iacr.org/2020/1604>`_.
"""
# pylint: disable=C0301 # Accommodate original format of published examples.
from __future__ import annotations
import doctest
from unittest import TestCase
from importlib import import_module
from itertools import product
from functools import reduce
import secrets
import hashlib
from parts import parts
from bitlist import bitlist

try:
    from circuitry import bit, bits, constants, synthesize
except: # pylint: disable=W0702
    # Support validation of docstrings in this script via its direct execution.
    import sys
    sys.path.append('./circuitry')
    from circuitry import bit, bits, constants, synthesize

@synthesize
def equal(x: bit, y: bit) -> bit:
    """
    Function that performs a simple single-bit equality calculation
    using logical and arithmetic operations.

    The synthesized :obj:`~circuit.circuit.circuit` object is introduced as
    an attribute of the function and can be evaluated on two bit values (as
    indicated by the function's type annotation).

    >>> equal.circuit.evaluate([[0], [1]])
    [[0]]

    Note that the function itself can still be invoked on its own in the usual
    manner if the supplied inputs are integers or :obj:`bit` instances. When
    the function is invoked in this way, the output of the function corresponds
    to its output type annotation.

    >>> equal(0, 1)
    0
    >>> b = equal(bit(0), bit(1))
    >>> isinstance(b, bit)
    True
    >>> int(b)
    0
    """
    return (x & y) | ((1 - x) & (1 - y))

@synthesize
def equals_iterative(xs: bits(8), ys: bits(8)) -> bit:
    """
    Function that employs an iterative approach and performs a bit vector
    equality calculation. This function invokes both the single-bit equality
    function defined above and the logical conjunction operation.

    >>> bs = [0, 1, 1, 0, 1, 0, 1, 0]
    >>> equals_iterative.circuit.evaluate([bs, bs])
    [[1]]
    >>> equals_iterative.circuit.evaluate([bs, list(reversed(bs))])
    [[0]]
    """
    z = 1
    for i in range(min(len(xs), len(ys))):
        z = z & equal(xs[i], ys[i])
    return z

@synthesize
def equals_functional(xs: bits(8), ys: bits(8)) -> bit:
    """
    Function that employs a functional approach and performs a bit vector
    equality calculation. This function invokes both the single-bit equality
    function defined above and the logical conjunction operation.

    >>> bs = [0, 1, 1, 0, 1, 0, 1, 0]
    >>> equals_functional.circuit.evaluate([bs, bs])
    [[1]]
    >>> equals_functional.circuit.evaluate([bs, list(reversed(bs))])
    [[0]]
    """
    es = [equal(x, y) for (x, y) in zip(xs, ys)]
    return reduce((lambda e0, e1: e0 & e1), es)

def add32(xs, ys):
    """
    Function that performs addition of two interegers that are represented
    as 32-bit vectors. This function is intended for use by the :obj:`sha256`
    function that implements SHA-256.

    Note that this is a helper function that is invoked by the :obj:`sha256`
    function. **It is not decorated** because **execution of** :obj:`sha256`
    **is what synthesizes the overall SHA-256 circuit**. The body of this
    function could have been inlined within the body of the :obj:`sha256`
    function without impacting the synthesis of the SHA-256 circuit.
    """
    (xs, ys) = (list(reversed(xs)), list(reversed(ys)))
    (x0, xt, y0, yt) = (xs[0], xs[1:], ys[0], ys[1:])
    carry = x0 & y0
    def combine(zs, xy):
        carry = zs.pop()
        add = xy[0] ^ xy[1] # Reuse via variable reduces number of synthesized gates.
        return zs + [add ^ carry, (xy[0] & xy[1]) | (add & carry)]

    return bits(list(reversed(
        [x0 ^ y0] + list(reduce(combine, zip(xt, yt), [carry]))[:-1]
    )))

def sha256_iteration(d_8_32s, m_64_8s):
    """
    Perform a single iteration of the SHA-256 digest computation on the current
    digest (consisting of 32 distinct bit vectors, each having 8 bits) based on
    a message portion (consisting of 64 distinct bit vectors, each having 8
    bits) to produce the next digest.

    Note that this is a helper function that is invoked by the :obj:`sha256`
    function. **It is not decorated** because **execution of** :obj:`sha256`
    **is what synthesizes the overall SHA-256 circuit**. The body of this
    function could have been inlined within the body of the :obj:`sha256`
    function without impacting the synthesis of the SHA-256 circuit.
    """
    # Table of constants (64 bit vectors each having 32 bits).
    table = [constants(list(bitlist(i, 32))) for i in [
        1116352408, 1899447441, 3049323471, 3921009573,  961987163, 1508970993, 2453635748, 2870763221,
        3624381080,  310598401,  607225278, 1426881987, 1925078388, 2162078206, 2614888103, 3248222580,
        3835390401, 4022224774,  264347078,  604807628,  770255983, 1249150122, 1555081692, 1996064986,
        2554220882, 2821834349, 2952996808, 3210313671, 3336571891, 3584528711,  113926993,  338241895,
        666307205,   773529912, 1294757372, 1396182291, 1695183700, 1986661051, 2177026350, 2456956037,
        2730485921, 2820302411, 3259730800, 3345764771, 3516065817, 3600352804, 4094571909,  275423344,
        430227734,   506948616,  659060556,  883997877,  958139571, 1322822218, 1537002063, 1747873779,
        1955562222, 2024104815, 2227730452, 2361852424, 2428436474, 2756734187, 3204031479, 3329325298
    ]]

    # Functions used during the hash computation.
    Ch = lambda x, y, z: (x & y) ^ ((~x) & z)
    Maj = lambda x, y, z: ((x & y) ^ (x & z)) ^ (y & z)
    Sigma_0 = lambda bs: ((bs >> {2}) ^ (bs >> {13})) ^ (bs >> {22})
    Sigma_1 = lambda bs: ((bs >> {6}) ^ (bs >> {11})) ^ (bs >> {25})
    sigma_0 = lambda bs: ((bs >> {7}) ^ (bs >> {18})) ^ (bs >> 3)
    sigma_1 = lambda bs: ((bs >> {17}) ^ (bs >> {19})) ^ (bs >> 10)

    w = [] # Message schedule.
    for j in range(16):
        w.append(m_64_8s[j*4] + m_64_8s[j*4+1] + m_64_8s[j*4+2] + m_64_8s[j*4+3])
    for j in range(16, 64):
        w.append(add32(add32(add32(sigma_1(w[j-2]), w[j-7]), sigma_0(w[j-15])), w[j-16]))

    wv = d_8_32s # Eight 32-bit working variables.
    for j in range(64):
        c = add32(add32(Ch(wv[4], wv[5], wv[6]), table[j]), w[j])
        t1 = add32(add32(wv[7], Sigma_1(wv[4])), c)
        t2 = add32(Sigma_0(wv[0]), Maj(wv[0], wv[1], wv[2]))
        wv = [add32(t1, t2), wv[0], wv[1], wv[2], add32(wv[3], t1), wv[4], wv[5], wv[6]]

    return [add32(d_8_32s[j], wv[j]) for j in range(8)] # Return next digest.

@synthesize
def sha256(message: bits(512)) -> bits(256):
    """
    Accept an appropriately padded bit vector of length 512 and compute a SHA-256 message
    digest as the output (represented as a bit vector having 256 bits).

    This SHA-256 algorithm conforms to the
    `FIPS 180-4 specification <https://www.tandfonline.com/doi/abs/10.1080/01611194.2012.687431>`_
    and expects inputs that are appropriately padded. The example below demonstrates
    how an appropriate input byte vector can be constructed for the synthesized
    circuit.

    >>> input_bytes = bytes([1, 2, 3])
    >>> input_padding = (
    ...     bytes([128] + ([0] * (55 - len(input_bytes)))) +
    ...     int(8 * len(input_bytes)).to_bytes(8, 'big')
    ... )

    When evaluating the synthesized circuit, the input must be converted into a bit vector.
    Note that the output consists of a list containing a single bit vector that has 256 bits.

    >>> [output_bits] = sha256.circuit.evaluate([
    ...     [
    ...         b
    ...         for byte in (input_bytes + input_padding)
    ...         for b in bitlist(byte, 8)
    ...     ]
    ... ])

    The output can be compared to a reference implementation.

    >>> bitlist(output_bits).hex() == hashlib.sha256(input_bytes).hexdigest()
    True
    """
    # Split input into list of vectors, each having 8 bits.
    message = [bits(p) for p in parts(message, length=8)]

    # Initial digest: eight bit vectors each having 32 bits.
    digest = [constants(list(bitlist(i, 32))) for i in [
        1779033703, 3144134277, 1013904242, 2773480762,
        1359893119, 2600822924,  528734635, 1541459225
    ]]

    # Perform hash computation for appropriate number of iterations (based on message length).
    for i in range(len(message) // 64):
        digest = sha256_iteration(digest, message[(i * 64) : (i * 64) + 64])

    # Flatten 8-bit vectors (each of 32 bits) into single 256-bit vector.
    return [b for bs in digest for b in bs]

class Test_circuitry(TestCase):
    """
    Tests involving published examples demonstrating the use of the library.
    """
    def test_exports(self):
        """
        Checks that the module exports the expected classes and functions.
        """
        module = import_module('circuitry.circuitry')
        self.assertTrue({
            'bit', 'constant', 'input', 'input_one', 'input_two', 'output', \
            'bits', 'constants', 'inputs', 'outputs', \
            'synthesize'
        }.issubset(module.__dict__.keys()))

    def test_example_equal(self):
        """
        Tests synthesis of a circuit for a simple single-bit equality
        function.
        """
        for (x, y) in product(*[[0, 1]] * 2):
            self.assertEqual(
                equal.circuit.evaluate([[x], [y]]),
                [[int(x == y)]]
            )

    def test_example_equals_iterative(self):
        """
        Tests synthesis of a circuit for a simple bit vector equality
        function.
        """
        vectors = product(*[[0, 1]] * 8)
        for (xs, ys) in product(vectors, vectors):
            self.assertEqual(
                equals_iterative.circuit.evaluate([xs, ys]),
                [[int(xs == ys)]]
            )

    def test_example_equals_functional(self):
        """
        Tests synthesis of a circuit for a simple bit vector equality
        function.
        """
        vectors = product(*[[0, 1]] * 8)
        for (xs, ys) in product(vectors, vectors):
            self.assertEqual(
                equals_functional.circuit.evaluate([xs, ys]),
                [[int(xs == ys)]]
            )

    def test_example_sha256(self):
        """
        Tests the circuit corresponding to an implementation of SHA-256 that was
        synthesized directly using the synthesis decorator for a specific length
        of input to the hash function.
        """
        # Perform a few tests for each length of input.
        for length in range(1, 56, 3):
            for _ in range(3):
                # Assemble input and padding.
                input_bytes = secrets.token_bytes(length)
                input_padding = \
                    bytes([128] + ([0] * (55 - length))) + \
                    int(8 * length).to_bytes(8, 'big')

                # Evaluate the synthesized circuit on the input bit vector.
                [output_bits] = sha256.circuit.evaluate([
                    [
                        b
                        for byte in (input_bytes + input_padding)
                        for b in bitlist(byte, 8)
                    ]
                ])

                # Compute circuit evaluation result to that of the built-in SHA-256
                # implementation.
                self.assertEqual(
                    bitlist(output_bits).hex(),
                    hashlib.sha256(input_bytes).hexdigest()
                )

    def test_example_sha256_resynthesis_per_input_length(self):
        """
        Tests multiple circuit variants corresponding to SHA-256, each synthesized
        programmatically for one of a collection of specific input length ranges.

        Synthesis of different SHA-256 circuits for different input lengths is
        necessary because unlike a general-purpose programming lanaguage with
        support fo iteration constructs (such as loops), circuits are fixed and
        can only operate on one length of input. Due to padding requirements, input
        lengths for SHA-256 come in exact multiples of 512; therefore, a distinct
        circuit must be synthesized for each multiple of 512.

        Note that while the :obj:`sha256` method is annotated for a specific input
        bit vector size, the tests below override this annotation by invoking
        circuit synthesis programmatically (thus supplying a different annotation).
        """
        # Perform a few tests for each length of input.
        for multiple in range(2, 6):
            # Synthesize the circuit for this particular length of input.
            c = synthesize(sha256, [bits(512 * multiple)], [bits(256)])

            # Run a test for the two boundary values of current input length range.
            length_minimum = (64 * (multiple - 1)) - 8
            for length in [length_minimum, length_minimum + 63]:
                # Assemble input and padding.
                input_bytes = secrets.token_bytes(length)
                input_padding = \
                    bytes([128] + ([0] * ((64 * multiple) - 9 - length))) + \
                    int(8 * length).to_bytes(8, 'big')

                # Evaluate the synthesized circuit on the input bit vector.
                [output_bits] = c.evaluate([
                    [
                        b
                        for byte in (input_bytes + input_padding)
                        for b in bitlist(byte, 8)
                    ]
                ])

                # Compute circuit evaluation result to that of the built-in SHA-256
                # implementation.
                self.assertEqual(
                    bitlist(output_bits).hex(),
                    hashlib.sha256(input_bytes).hexdigest()
                )

# Always invoke the doctests in this module.
doctest.testmod()



"""
Embedded domain-specific combinator library for assembling abstract definitions
of logical circuits and for automatically synthesizing circuits from those
definitions.

This library makes it possible both to construct circuits **programmatically** (see
the documentation for the :obj:`bit` class) and to synthesize circuits **in an
automated manner** from Python function definitions (see the documentation for
the :obj:`synthesize` function).
"""
from __future__ import annotations
from typing import Sequence, Union, Optional, Callable
import doctest
import inspect
from parts import parts
from circuit import op, gate, circuit, signature

class bit:
    """
    Class for representing an abstract bit. Such a bit can be interpreted
    concretely as a value, but it is also used to keep track of relationships
    between operators and to represent the individual wires within a circuit
    that is programmatically built up from gates that correspond to those
    operators.

    **The documentation for this class describes how circuits can be
    synthesized programmatically using the methods made available by this
    class. Consult the documentation for the** :obj:`synthesize` **function
    to learn how circuits can be synthesized automatically from Python function
    definitions.**

    When constructing a circuit programmatically, it is first necessary to
    introduce a new circuit object and to designate this object as the circuit
    that is being constructed.

    >>> c = circuit()
    >>> bit.circuit(c)

    Subsequently, it is possible to instantiate subclasses of :obj:`bit` that
    serve specific roles (such as :obj:`input`, which represents an input value
    into the overall circuit being constructed). An operator method such as
    :obj:`bit.and_` can be used to operate on :obj:`bit` instances (and, at the
    same time to introduce a gate into the circuit).

    >>> b = output(input(1).and_(input(1)))

    At this point, constructed circuit can be retrieved and evaluated on a
    vector of bit values (wherein each value is represented using the integer
    ``0`` or the integer ``1``).

    >>> b.value == bit.circuit().evaluate([1, 1])[0]
    True

    It is possible to add custom hook functions that are applied whenever an
    operator is introduced using the :obj:`bit.hook_operation` method.

    >>> def make_hook_that_prints_created_gates(bit_):
    ...     def hook(o, v, *args):
    ...         print('created gate with operation "' + o.name() + '"')
    ...         return bit_.constructor(*args)(v, bit_.gate(o, [a.gate for a in args]))
    ...     return hook
    >>> bit.hook_operation(make_hook_that_prints_created_gates(bit))
    >>> bit.circuit(circuit())
    >>> b = output(input(0).and_(input(1)).not_())
    created gate with operation "and"
    created gate with operation "not"

    Note that a hook must be removed if its effects are not desired when subsequent
    circuits are constructed.

    >>> bit.hook_operation()
    """
    _circuit = None
    _hook_operation = None

    def __init__(self: bit, value: int, gate_: Optional[gate] = None):
        """
        Create an instance with the specified value and (if one is supplied)
        designate an associate gate object.
        """
        self.value = value

        if bit._circuit is not None:
            self.gate = bit._circuit.gate() if gate_ is None else gate_

    @staticmethod
    def circuit(circuit_: Optional[circuit] = None) -> Optional[circuit]:
        """
        Designate the circuit object that is under construction. Any invocation
        of the :obj:`bit` constructor adds a gate to the circuit designated
        using this method.

        >>> c = circuit()
        >>> bit.circuit(c)
        >>> b = output(input(1).and_(input(1)))
        >>> b.value == bit.circuit().evaluate([1, 1])[0]
        True

        Invoking this method with no argument retrieves *and removes* the
        circuit object (leaving no designated circuit object). Note that
        because of the invocation of this method at the end of the above
        example, the invocation below returns ``None``.

        >>> bit.circuit() is None
        True
        """
        if circuit_ is not None:
            bit._circuit = circuit_
            return None

        if bit._circuit is not None:
            bit._circuit.prune_and_topological_sort_stable()
            c = bit._circuit
            bit._circuit = None
            return c

        return None

    @staticmethod
    def hook_operation(hook: Optional[Callable] = None):
        """
        Assign a function that is invoked whenever the :obj:`bit.operation` method
        is used to create a new instance of :obj:`bit`.

        >>> def make_hook_that_prints_created_gates(bit_):
        ...     def hook(o, v, *args):
        ...         print('created gate with operation "' + o.name() + '"')
        ...         return bit_.constructor(*args)(v, bit_.gate(o, [a.gate for a in args]))
        ...     return hook
        >>> bit.hook_operation(make_hook_that_prints_created_gates(bit))
        >>> bit.circuit(circuit())
        >>> b = output(input(0).and_(input(1)).not_())
        created gate with operation "and"
        created gate with operation "not"
        >>> b.value == bit.circuit().evaluate([0, 0])[0]
        True

        If a hook function returns ``None``, then the default instance of :obj:`bit`
        is returned when an operation is applied.

        >>> bit.hook_operation(lambda o, v, *args: None)
        >>> bit.circuit(circuit())
        >>> b = output(input(1).and_(input(1)))
        >>> b.value == bit.circuit().evaluate([1, 1])[0]
        True

        It is only possible to assign one hook function at a time. A hook must
        be removed if its effects are not desired when subsequent circuits are
        constructed.

        >>> bit.hook_operation()
        """
        bit._hook_operation = hook

    @staticmethod
    def operation(o: Callable, *args) -> bit:
        """
        Apply the supplied operation method to zero or more :obj:`bit` arguments.
        Gate operations are represented using instances of the
        :obj:`~logical.logical.logical` class exported by the
        `logical <https://pypi.org/project/logical/>`_ library. This module
        indirectly imports the :obj:`~logical.logical.logical` class via the
        :obj:`~circuit.circuit.op` synonym defined in the
        `circuit <https://pypi.org/project/circuit/>`_ library, enabling the
        more concise syntax used in the example below.

        >>> bit.circuit(circuit())
        >>> b = output(bit.operation(op.and_, input(1), input(1)))
        >>> b.value == bit.circuit().evaluate([1, 1])[0]
        True

        Arguments that are instances of :obj:`output` are not permitted.

        >>> bit.circuit(circuit())
        >>> b0 = output(input(0).not_())
        >>> b1 = b0.not_()
        Traceback (most recent call last):
          ...
        TypeError: cannot supply an output as an argument to an operation
        >>> _ = bit.circuit() # Remove designated circuit.
        """
        # Ensure second argument is a `bit`.
        args = list(args)
        if len(args) == 2:
            args[1] = constant(args[1]) if isinstance(args[1], int) else args[1]

        # Ensure none of the arguments are outputs.
        for a in args:
            if isinstance(a, output):
                raise TypeError(
                    "cannot supply an output as an argument to an operation"
                )

        # Compute the value of the result of the operation on the arguments.
        v = o(*[a.value for a in args])

        # Return output from hook if it exists and if
        # it returns an output.
        if bit._hook_operation is not None:
            r = bit._hook_operation(o, v, *args)
            if r is not None:
                return r

        return bit.constructor(*args)(v, bit.gate(o, [a.gate for a in args]))

    @staticmethod
    def constructor(b1: bit, b2: Optional[bit] = None) -> type:
        """
        Return the constructor to use for instantiating a :obj:`bit` object.
        This method is used by the :obj:`bit.operation` and :obj:`bits.from_byte`
        methods. It is provided primarily to aid in the implementation of custom
        hooks that can be set using the :obj:`bit.hook_operation` method.
        """
        return bit

    @staticmethod
    def gate(operation: op, igs: Sequence[gate]) -> Optional[gate]:
        """
        Add a gate to the designated circuit object that is under construction.
        This method is primarily provided to aid in the implementation of custom
        hooks that can be set using the :obj:`bit.hook_operation` method.
        Gate operations are represented using instances of the
        :obj:`~logical.logical.logical` class exported by the
        `logical <https://pypi.org/project/logical/>`_ library (which is indirectly
        imported into this module as the :obj:`~circuit.circuit.op` synonym
        from the `circuit <https://pypi.org/project/circuit/>`_ library).

        >>> def make_hook(bit_):
        ...     def hook(o, v, *args):
        ...         return bit_.constructor(*args)(v, bit_.gate(o, [a.gate for a in args]))
        ...     return hook
        >>> bit.hook_operation(make_hook(bit))
        >>> bit.circuit(circuit())
        >>> b = output(input(0).and_(input(0)))
        >>> b.value == bit.circuit().evaluate([0, 0])[0]
        True
        """
        return (
            bit._circuit.gate(operation, igs) \
            if bit._circuit is not None else \
            None
        )

    def __int__(self: bit) -> int:
        """
        Convert this :obj:`bit` instance into the integer representation of
        its value.

        >>> int(bit(1))
        1
        """
        return self.value

    def id_(self: bit) -> bit:
        """
        Logical operation for an individual :obj:`bit` instance.

        >>> results = []
        >>> for x in [0, 1]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).id_())
        ...     results.append(int(b) == bit.circuit().evaluate([x])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.id_, self)

    def not_(self: bit) -> bit:
        """
        Logical operation for an individual :obj:`bit` instance.

        >>> results = []
        >>> for x in [0, 1]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).not_())
        ...     results.append(int(b) == bit.circuit().evaluate([x])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.not_, self)

    def __invert__(self: bit) -> bit:
        """
        Logical operation for an individual :obj:`bit` instance.

        >>> results = []
        >>> for x in [0, 1]:
        ...     bit.circuit(circuit())
        ...     b = output(~input(x))
        ...     results.append(int(b) == bit.circuit().evaluate([x])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.not_, self)

    def __rsub__(self: bit, other: bit) -> bit:
        """
        Arithmetic operation involving an individual :obj:`bit` instance that
        is often used to represent the logical negation operation. This method
        is provided primarily as a convenience enabling use of the syntax
        ``1 - b`` to represent logical negation of a bit value ``b``; it
        **does not enable subtraction of one** :obj:`bit` **instance from
        another**.

        >>> results = []
        >>> for x in [0, 1]:
        ...     bit.circuit(circuit())
        ...     b = output(1 - input(x))
        ...     results.append(int(b) == bit.circuit().evaluate([x])[0])
        >>> all(results)
        True
        >>> 2 - input(0)
        Traceback (most recent call last):
          ...
        ValueError: can only subtract a bit from the integer 1
        """
        if other == 1:
            return bit.operation(op.not_, self)
        raise ValueError('can only subtract a bit from the integer 1')

    def and_(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).and_(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.and_, self, other)

    def __and__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x) & input(y))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.and_, self, other)

    def __rand__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> b = 0 & constant(1)
        >>> b.value
        0
        """
        return self & (constant(other) if isinstance(other, int) else other)

    def nimp(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).nimp(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.nimp_, self, other)

    def nimp_(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).nimp_(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.nimp_, self, other)

    def __gt__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x) > input(y))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return self.nimp(other)

    def nif(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).nif(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.nif_, self, other)

    def nif_(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).nif_(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.nif_, self, other)

    def __lt__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x) < input(y))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return self.nif(other)

    def xor(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).xor(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.xor_, self, other)

    def xor_(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).xor_(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.xor_, self, other)

    def __xor__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x) ^ input(y))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.xor_, self, other)

    def __rxor__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> b =  1 ^ constant(0)
        >>> b.value
        1
        """
        return self ^ (constant(other) if isinstance(other, int) else other)

    def or_(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).or_(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.or_, self, other)

    def __or__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x) | input(y))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.or_, self, other)

    def __ror__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> b = 1 | constant(0)
        >>> b.value
        1
        """
        return self | (constant(other) if isinstance(other, int) else other)

    def nor(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).nor(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.nor_, self, other)

    def nor_(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).nor_(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.nor_, self, other)

    def __mod__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x) % input(y))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.nor_, self, other)

    def xnor(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).xnor(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.xnor_, self, other)

    def xnor_(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).xnor_(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.xnor_, self, other)

    def __eq__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x) == input(y))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.xnor_, self, other)

    def if_(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).if_(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.if_, self, other)

    def __ge__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x) >= input(y))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.if_, self, other)

    def imp(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).imp(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.imp_, self, other)

    def imp_(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).imp_(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.imp_, self, other)

    def __le__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x) <= input(y))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.imp_, self, other)

    def nand(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).nand(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.nand_, self, other)

    def nand_(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x).nand_(input(y)))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.nand_, self, other)

    def __matmul__(self: bit, other: bit) -> bit:
        """
        Logical operation for individual :obj:`bit` instances.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     b = output(input(x) @ input(y))
        ...     results.append(int(b) == bit.circuit().evaluate([x, y])[0])
        >>> all(results)
        True
        """
        return bit.operation(op.nand_, self, other)

class constant(bit):
    """
    Instance of a :obj:`bit` that is designated as a constant value.

    >>> constant(1).value
    1

    When a :obj:`constant` instance is introduced during circuit construction,
    a gate with a nullary operator (*i.e.*, one of the two appearing in
    :obj:`~logical.logical.logical.nullary`) is added to the circuit.

    >>> bit.circuit(circuit())
    >>> _ = output(constant(0))
    >>> c = bit.circuit()
    >>> c.count()
    2
    >>> c.evaluate([])
    [0]
    """
    def __init__(self: bit, value: int):
        """Instantiate an instance that is designated as a constant input."""
        self.value = value

        if bit._circuit is not None:
            self.gate = bit._circuit.gate(op.nf_ if self.value == 0 else op.nt_)

class input(bit):
    """
    Instance of a :obj:`bit` that is designated as a variable input. When an
    :obj:`input` instance is introduced during circuit construction, a gate
    is added to the circuit that is explicitly marked as an input gate (as
    defined in the `circuit <https://pypi.org/project/circuit/>`_ library).

    >>> b0 = output(input(1).not_())
    >>> b0.value
    0
    """
    def __init__(self: bit, value: int):
        """Instantiate an instance that is designated as a variable input."""
        self.value = value

        if bit._circuit is not None:
            self.gate = bit._circuit.gate(op.id_, is_input=True)

class input_one(input):
    """
    Instance of a :obj:`bit` that is designated as a variable input from one source.
    """

class input_two(input):
    """
    Instance of a :obj:`bit` that is designated as a variable input from a second source
    """

class output(bit):
    """
    Instance of a :obj:`bit` that is designated as an output. When an
    :obj:`output` instance is introduced during circuit construction, a gate is
    added to the circuit that is explicitly marked as an output gate (as defined
    in the `circuit <https://pypi.org/project/circuit/>`_ library).

    >>> bit.circuit(circuit())
    >>> b0 = input(0).not_()
    >>> b1 = output(b0.not_())
    >>> [b0.value, b1.value]
    [1, 0]
    >>> bit.circuit().evaluate([1])
    [1]

    It is not possible to apply an operation to an :obj:`output` instance. Any
    attempt to do so raises an exception (see implementation of :obj:`bit.operation`
    for more details).

    >>> bit.circuit(circuit())
    >>> b0 = output(input(0).not_())
    >>> b1 = b0.not_()
    Traceback (most recent call last):
      ...
    TypeError: cannot supply an output as an argument to an operation

    If the value represented by a :obj:`bit` instance must be both an argument
    into another operation and an output, an :obj:`bit.id_` operation can be
    introduced.

    >>> bit.circuit(circuit())
    >>> b0 = input(1).not_()
    >>> b1 = output(b0.id_())
    >>> b2 = output(b0.not_())
    >>> [b0.value, b1.value, b2.value]
    [0, 0, 1]
    >>> bit.circuit().evaluate([0])
    [1, 0]
    """
    def __init__(self: bit, b: bit):
        """
        Instantiate a bit that is designated as an output.
        """
        self.value = b.value

        if bit._circuit is not None:
            self.gate = bit._circuit.gate(op.id_, [b.gate], is_output=True)

class bits_type(int): # pylint: disable=R0903
    """
    Class for representing an input or output type of a function decorated
    for automated synthesis. This class is employed internally within the
    :obj:`bits` constructor implementation and is not intended to be used
    directly. It is not exported.
    """
    def __str__(self):
        """
        Return representation that reflects the meaning and the manner of
        construction of this instance.

        >>> str(bits_type(32))
        'bits(32)'
        """
        return 'bits(' + str(int(self)) + ')'

    def __repr__(self):
        """
        Return representation that reflects the meaning and the manner of
        construction of this instance.

        >>> bits_type(32)
        bits(32)
        """
        return 'bits(' + str(int(self)) + ')'

class bits(list):
    """
    Class for representing a *bit vector* (*i.e.*, a list of abstract :obj:`bit`
    instances).

    >>> bs = bits([constant(1)] * 3)
    >>> [b.value for b in bs]
    [1, 1, 1]
    """
    def __new__(cls, argument=None) -> bits:
        """
        Instantiate bit vector object given the supplied argument.
        """
        return bits_type(argument)\
            if isinstance(argument, int) else\
            list.__new__(cls, argument)

    @staticmethod
    def from_byte(byte_: int, constructor=bit) -> bits:
        """
        Convert an integer representing a single byte into a corresponding bit
        vector.

        >>> [b.value for b in bits.from_byte(255)]
        [1, 1, 1, 1, 1, 1, 1, 1]
        """
        return bits([
            constructor(bit_)
            for bit_ in reversed([(byte_ >> i) % 2 for i in range(8)])
        ])

    @staticmethod
    def from_bytes(bytes_, constructor=bit) -> bits:
        """
        Convert a vector of bytes into a corresponding bit vector.

        >>> [b.value for b in bits.from_bytes(bytes([255]))]
        [1, 1, 1, 1, 1, 1, 1, 1]
        >>> [b.value for b in bits.from_bytes(bytes([11, 0]))]
        [0, 0, 0, 0, 1, 0, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0]
        """
        return bits([
            bit_
            for byte_ in bytes_
            for bit_ in bits.from_byte(byte_, constructor)
        ])

    @staticmethod
    def zeros(n: int) -> bits:
        """
        Create a vector (of the specified length) of constant zero bits.

        >>> bit.circuit(circuit())
        >>> xs = bits.zeros(3)
        >>> ys = outputs(xs.not_())
        >>> [y.value for y in ys]
        [1, 1, 1]
        >>> _ = bit.circuit() # Remove designated circuit.
        """
        return bits([constant(0)]*n)

    def __int__(self: bits) -> int:
        """
        Convert a bit vector into an integer.

        >>> bit.circuit(circuit())
        >>> xs = constants([0, 0, 0])
        >>> ys = outputs(xs.not_())
        >>> int(ys)
        7
        >>> _ = bit.circuit() # Remove designated circuit.
        """
        return sum(int(b)*(2**i) for (i, b) in zip(range(len(self)), reversed(self)))

    def not_(self: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for x in [0, 1]:
        ...     bit.circuit(circuit())
        ...     xs = inputs([x, x, x])
        ...     ys = outputs(xs.not_())
        ...     ns = [int(y) for y in ys]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x]))
        >>> all(results)
        True
        """
        return bits([x.not_() for x in self])

    def __invert__(self: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for x in [0, 1]:
        ...     bit.circuit(circuit())
        ...     xs = inputs([x, x, x])
        ...     ys = outputs(~xs)
        ...     ns = [int(y) for y in ys]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x]))
        >>> all(results)
        True
        """
        return bits([x.not_() for x in self])

    def and_(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.and_(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.and_(y) for (x, y) in zip(self, other)])

    def __and__(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs & ys)
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.and_(y) for (x, y) in zip(self, other)])

    def nimp(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.nimp(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.nimp_(y) for (x, y) in zip(self, other)])

    def nimp_(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.nimp_(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.nimp_(y) for (x, y) in zip(self, other)])

    def __gt__(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs > ys)
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.nimp_(y) for (x, y) in zip(self, other)])

    def nif(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.nif(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.nif_(y) for (x, y) in zip(self, other)])

    def nif_(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.nif_(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.nif_(y) for (x, y) in zip(self, other)])

    def __lt__(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs < ys)
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.nif_(y) for (x, y) in zip(self, other)])

    def xor(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.xor(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.xor_(y) for (x, y) in zip(self, other)])

    def xor_(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.xor_(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.xor_(y) for (x, y) in zip(self, other)])

    def __xor__(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs ^ ys)
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.xor_(y) for (x, y) in zip(self, other)])

    def or_(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.or_(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.or_(y) for (x, y) in zip(self, other)])

    def __or__(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs | ys)
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.or_(y) for (x, y) in zip(self, other)])

    def nor(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.nor(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.nor_(y) for (x, y) in zip(self, other)])

    def nor_(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.nor_(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.nor_(y) for (x, y) in zip(self, other)])

    def __mod__(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs % ys)
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.nor_(y) for (x, y) in zip(self, other)])

    def xnor(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.xnor(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.xnor_(y) for (x, y) in zip(self, other)])

    def xnor_(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.xnor_(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.xnor_(y) for (x, y) in zip(self, other)])

    def __eq__(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs == ys)
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.xnor_(y) for (x, y) in zip(self, other)])

    def if_(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.if_(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.if_(y) for (x, y) in zip(self, other)])

    def __ge__(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs >= ys)
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.if_(y) for (x, y) in zip(self, other)])

    def imp(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.imp(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.imp_(y) for (x, y) in zip(self, other)])

    def imp_(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.imp_(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.imp_(y) for (x, y) in zip(self, other)])

    def __le__(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs <= ys)
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.imp_(y) for (x, y) in zip(self, other)])

    def nand(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.nand(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.nand_(y) for (x, y) in zip(self, other)])

    def nand_(self: bits, other: bits) -> bits:
        """
        Logical operation for bit vectors.

        >>> results = []
        >>> for (x, y) in [(0, 0), (0, 1), (1, 0), (1, 1)]:
        ...     bit.circuit(circuit())
        ...     (xs, ys) = (inputs([x, x, x]), inputs([y, y, y]))
        ...     zs = outputs(xs.nand_(ys))
        ...     ns = [int(z) for z in zs]
        ...     c = bit.circuit()
        ...     results.append(ns == c.evaluate([x, x, x, y, y, y]))
        >>> all(results)
        True
        """
        return bits([x.nand_(y) for (x, y) in zip(self, other)])

    def __rshift__(self: bits, other: Union[int, set]) -> bits:
        """
        Overloaded operator for performing rotation and shift operations on
        bit vectors. When the second parameter is an integer, a shift operation
        is performed. When the second parameter is a set containing a single
        integer, a rotation operation is performed.

        >>> bs = bits(map(bit, [1, 1, 1, 1, 0, 0, 0, 0]))
        >>> bs = bs >> 3
        >>> [b.value for b in bs]
        [0, 0, 0, 1, 1, 1, 1, 0]
        >>> bs = bits(map(bit, [0, 0, 0, 0, 1, 1, 1, 1]))
        >>> bs = bs >> {3}
        >>> [b.value for b in bs]
        [1, 1, 1, 0, 0, 0, 0, 1]
        """
        if isinstance(other, set) and isinstance(list(other)[0], int): # Rotation.
            quantity = list(other)[0]
            return bits(self[len(self)-quantity:]) ** bits(self[0:len(self)-quantity])
        else: # Shift
            return bits([constant(0)]*other) ** bits(self[0:len(self)-other])

    def __lshift__(self: bits, other: int) -> bits:
        """
        Overloaded operator for performing shift operations on bit vectors.

        >>> bs = bits(map(bit, [1, 1, 1, 1, 0, 0, 0, 0]))
        >>> bs = bs << 3
        >>> [b.value for b in bs]
        [1, 0, 0, 0, 0, 0, 0, 0]
        """
        return bits(self[other:]) ** bits([constant(0) for _ in range(other)])

    def __truediv__(self: bits, other: [Union, set, list]) -> Sequence[bits]:
        """
        Overloaded operator for splitting a bit vector into a collection of
        smaller bit vectors. Three operations are possible depending on the
        type of the second parameter.

        When the second parameter is an integer, this instance is split into
        that number of bit vectors. If the size of the vector is not an exact
        multiple of the integer, the parts in the result may not all be the
        same length.

        >>> bs = bits(map(bit, [1, 1, 1, 1, 0, 0, 0, 0]))
        >>> bss = list(bs / 2)
        >>> ([b.value for b in bss[0]], [b.value for b in bss[1]])
        ([1, 1, 1, 1], [0, 0, 0, 0])
        >>> bs = bits(map(bit, [1, 1, 1, 1, 0, 0, 0, 0]))
        >>> [len(bs) for bs in list(bs / 3)]
        [2, 3, 3]

        When the second parameter is a set containing a single integer, this
        instance is split into parts of the size specified by that integer.

        >>> bs = bits(map(bit, [1, 1, 1, 1, 0, 0, 0, 0]))
        >>> bss = list(bs / {2})
        >>> [[b.value for b in bs] for bs in bss]
        [[1, 1], [1, 1], [0, 0], [0, 0]]
        >>> bs = bits(map(bit, [1, 1, 1, 1, 0, 0, 0, 0]))
        >>> bss = list(bs / {1})
        >>> [[b.value for b in bs] for bs in bss]
        [[1], [1], [1], [1], [0], [0], [0], [0]]

        When the second parameter is a list of integers, this instance is split
        into parts the sizes of which correspond to the entries (in order) of
        the list of integers.

        >>> bs = bits(map(bit, [1, 1, 1, 1, 0, 0, 0, 0]))
        >>> bss = list(bs / [1, 3, 4])
        >>> [[b.value for b in bs] for bs in bss]
        [[1], [1, 1, 1], [0, 0, 0, 0]]
        """
        if isinstance(other, list) and len(other) > 0 and isinstance(other[0], int):
            return map(bits, parts(self, length=other)) # Sequence of lengths.
        elif isinstance(other, set) and len(other) == 1 and isinstance(list(other)[0], int):
            return self / (len(self)//list(other)[0]) # Parts of length `other`.
        else:
            return map(bits, parts(self, other)) # Number of parts is `other`.

    def __add__(self: bits, other: Union[bits, Sequence[int]]) -> bits:
        """
        Overloaded operator for concatenating bit vectors.

        >>> bs = bits(map(bit, [1, 1])) + bits(map(bit, [0, 0]))
        >>> [b.value for b in bs]
        [1, 1, 0, 0]
        """
        result = list(self)
        result.extend(list(other))
        return bits(result)

    def __pow__(self: bits, other: Union[bits, Sequence[int]]) -> bits:
        """
        Overloaded operator for concatenating bit vectors.

        >>> bs = bits(map(bit, [1, 1])) ** bits(map(bit, [0, 0]))
        >>> [b.value for b in bs]
        [1, 1, 0, 0]
        """
        return self + other

def constants(l: Sequence[int]) -> bits:
    """
    Synonym for concisely constructing a vector of :obj:`bit` objects that
    are designated as constants.

    >>> bit.circuit(circuit())
    >>> xs = constants([0, 0, 0])
    >>> ys = outputs(xs.not_())
    >>> int(ys)
    7
    >>> _ = bit.circuit() # Remove designated circuit.
    """
    return bits(map(constant, l))

def inputs(l: Sequence[int]) -> bits:
    """
    Synonym for concisely constructing a vector of :obj:`bit` objects that
    are designated as inputs.

    >>> results = []
    >>> bit.circuit(circuit())
    >>> xs = inputs([1, 1, 1])
    >>> ys = outputs(xs.not_())
    >>> ns = [int(y) for y in ys]
    >>> c = bit.circuit()
    >>> ns == c.evaluate([1, 1, 1])
    True
    """
    return bits(map(input, l))

def outputs(l: Sequence[int]) -> bits:
    """
    Synonym for concisely constructing a vector of :obj:`bit` objects that
    are designated as outputs.

    >>> bit.circuit(circuit())
    >>> xs = bits.zeros(3)
    >>> ys = outputs(xs.not_())
    >>> [y.value for y in ys]
    [1, 1, 1]
    >>> _ = bit.circuit() # Remove designated circuit.
    """
    return bits(map(output, l))

def synthesize(function: Callable, in_type=None, out_type=None) -> Callable:
    """
    Decorator for automatically synthesizing a circuit from a function that
    takes only :obj:`bit` and/or :obj:`bits` objects as its arguments and
    returns an output of type :obj:`bit` or :obj:`bits` (or a tuple or list
    thereof).

    In the example below, a function ``equal`` is defined that determines
    whether two bits are equivalent. The use of this decorator causes a
    circuit that implements this function to be constructed.

    >>> @synthesize
    ... def equal(x: bit, y: bit) -> bit:
    ...     return (x & y) | ((1 - x) & (1 - y))

    The synthesized :obj:`~circuit.circuit.circuit` object is introduced as
    an attribute of the function and can be evaluated on two bit values (as
    indicated by the function's type annotation).

    >>> [equal.circuit.evaluate([[x], [y]]) for x in (0, 1) for y in (0, 1)]
    [[[1]], [[0]], [[0]], [[1]]]

    Note that the function itself can still be invoked on its own in the usual
    manner if the supplied inputs are integers or :obj:`bit` instances. When
    the function is invoked in this way, the output of the function corresponds
    to its output type annotation.

    >>> equal(0, 1)
    0
    >>> b = equal(bit(0), bit(1))
    >>> isinstance(b, bit)
    True
    >>> int(b)
    0

    This decorator can also be applied to functions that are defined
    explicitly as operating on bit vectors (in the form of :obj:`bits`
    objects).

    >>> @synthesize
    ... def conjunction(xy: bits(2)) -> bits(2):
    ...     return (xy[0], xy[0] & xy[1])
    >>> [conjunction.circuit.evaluate([[x, y]]) for x in (0, 1) for y in (0, 1)]
    [[[0, 0]], [[0, 0]], [[1, 0]], [[1, 1]]]

    If the decorated function returns multiple outputs, the output type annotation
    should indicate this.

    >>> @synthesize
    ... def swap(x: bit, y: bit) -> (bit, bit):
    ...     return (y, x)
    >>> [swap.circuit.evaluate([[x], [y]]) for x in (0, 1) for y in (0, 1)]
    [[[0], [0]], [[1], [0]], [[0], [1]], [[1], [1]]]

    Functions to which this decorator is applied must have valid type annotations
    that specify the lengths of the input and output bit vectors.

    >>> @synthesize
    ... def equal(x, y):
    ...     return x & y
    Traceback (most recent call last):
      ...
    ValueError: function must have a type annotation for every argument
    >>> @synthesize
    ... def equal(x: bit, y: bit):
    ...     return x & y
    Traceback (most recent call last):
      ...
    ValueError: function must have an output type annotation
    >>> @synthesize
    ... def equal(x: 'ABC', y: 'ABC') -> bit:
    ...     return x & y
    Traceback (most recent call last):
      ...
    TypeError: input type annotations must be specified using bit or bits
    >>> @synthesize
    ... def swap(x: bit, y: bit) -> ('ABC', 'ABC'):
    ...     return (y, x)
    Traceback (most recent call last):
      ...
    TypeError: output type annotation components must be specified using bit or bits
    >>> @synthesize
    ... def swap(x: bit, y: bit) -> 'ABC':
    ...     return (y, x)
    Traceback (most recent call last):
      ...
    TypeError: output type must be specified using bit/bits, or a list/tuple thereof

    If an exception occurs during the execution (for the purpose of circuit synthesis)
    of the decorated function, then synthesis will fail.

    >>> @synthesize
    ... def equal(x: bit, y: bit) -> bit:
    ...     return 1 / 0 # Run-time error.
    Traceback (most recent call last):
      ...
    RuntimeError: automated circuit synthesis failed

    To support programmatic synthesis (*e.g.*, of circuit variants of many different
    sizes from the same function definition), this function can accept input and output
    type information via two optional parameters.

    >>> def conjunction(xy):
    ...     return (xy[0], xy[0] & xy[1])
    >>> c = synthesize(conjunction, {'xy': bits(2)}, bits(2))
    >>> [c.evaluate([[x, y]]) for x in (0, 1) for y in (0, 1)]
    [[[0, 0]], [[0, 0]], [[1, 0]], [[1, 1]]]

    >>> def swap(x: bit, y: bit) -> (bit, bit):
    ...     return (y, x)
    >>> c = synthesize(swap, (bit, bit), (bit, bit))
    >>> [c.evaluate([[x], [y]]) for x in (0, 1) for y in (0, 1)]
    [[[0], [0]], [[1], [0]], [[0], [1]], [[1], [1]]]

    When synthesizing programmatically, the input type information must be supplied
    either (1) as a dictionary that maps input type parameter names to types or (2)
    as a tuple or list (the length of which matches the number of parameters). The
    output type information must follow the same conventions as an output type
    annotation.

    >>> def equal(x, y):
    ...     return (x & y) | ((1 - x) & (1 - y))
    >>> c = synthesize(equal, (bit, bit), bit)
    >>> [c.evaluate([[x], [y]]) for x in (0, 1) for y in (0, 1)]
    [[[1]], [[0]], [[0]], [[1]]]
    >>> def swap(xy):
    ...     return [xy[1], xy[0]]
    >>> c = synthesize(swap, bits(2), bits(2))
    >>> [c.evaluate([[x, y]]) for x in (0, 1) for y in (0, 1)]
    [[[0, 0]], [[1, 0]], [[0, 1]], [[1, 1]]]

    If the supplied type information does not have the correct type or is incomplete,
    an exception is raised.

    >>> c = synthesize(equal, (bit, bit))
    Traceback (most recent call last):
      ...
    ValueError: must include input and output types when supplying type information via parameters
    >>> c = synthesize(equal, bits(2), bit)
    Traceback (most recent call last):
      ...
    ValueError: number of input type components does not match number of function arguments
    >>> c = synthesize(equal, ('ABC', 'ABC'), bit)
    Traceback (most recent call last):
      ...
    TypeError: input type components must be specified using bit or bits
    >>> c = synthesize(equal, 'bits(2)', bit)
    Traceback (most recent call last):
      ...
    TypeError: input type must be specified using bit/bits, or a dict/list/tuple thereof
    >>> c = synthesize(equal, {'x': 'ABC', 'y': 'ABC'}, bit)
    Traceback (most recent call last):
      ...
    TypeError: input type components must be specified using bit or bits
    >>> c = synthesize(equal, (bit, bit), 'ABC')
    Traceback (most recent call last):
      ...
    TypeError: output type must be specified using bit/bits, or a list/tuple thereof
    >>> c = synthesize(equal, (bit, bit), ('ABC', bit))
    Traceback (most recent call last):
      ...
    TypeError: output type components must be specified using bit or bits
    """
    # pylint: disable=R0912,R0915

    # Assume the type information is supplied via type annotations for the function until
    # it is determined otherwise.
    type_supplied_via_annotation = True

    # For forward-compatibility with PEP 563.
    eval_ = lambda a: eval(a) if isinstance(a, str) else a # pylint: disable=W0123

    # If the type information is supplied via parameters, then both input and output
    # types must be supplied.
    if (in_type is None and out_type is not None) or (in_type is not None and out_type is None):
        raise ValueError(
            "must include input and output types when supplying type information via parameters"
        )

    # If the type information is supplied via parameters, then both input and output
    # types must be supplied.
    if (in_type is not None) and (out_type is not None):
        type_supplied_via_annotation = False

        # Ensure that the input type is specified in a valid way.
        if in_type is bit or isinstance(in_type, bits_type):
            in_type = (in_type,)
        elif isinstance(in_type, (tuple, list)):
            if not all(t is bit or isinstance(t, bits_type) for t in in_type):
                raise TypeError("input type components must be specified using bit or bits")
        elif isinstance(in_type, dict):
            if not all(t is bit or isinstance(t, bits_type) for t in in_type.values()):
                raise TypeError("input type components must be specified using bit or bits")
        else:
            raise TypeError(
                "input type must be specified using bit/bits, or a dict/list/tuple thereof"
            )

        # If the type information is supplied via parameters, ensure that the input type
        # information is converted into a dictionary that maps the function's input variables
        # to the corresponding type information (*i.e.*, if the user supplied a tuple or list
        # and not a dictionary for the input type information).
        if isinstance(in_type, (tuple, list)):
            if len(in_type) != len(inspect.getfullargspec(function).args):
                raise ValueError(
                    "number of input type components does not match number of function arguments"
                )
            in_type = dict(zip(inspect.getfullargspec(function).args, in_type))

        # Ensure that the output type is specified in a valid way.
        if out_type is bit or isinstance(out_type, bits_type):
            out_type = (out_type,)
        elif isinstance(out_type, (tuple, list)):
            if not all(t is bit or isinstance(t, bits_type) for t in out_type):
                raise TypeError("output type components must be specified using bit or bits")
        else:
            raise TypeError(
                "output type must be specified using bit/bits, or a list/tuple thereof"
            )

    # If the type information is supplied the function's annotation, ensure that its structure
    # is valid (and extract it for later use).
    if type_supplied_via_annotation:
        # Ensure that every one of the function's arguments and its output has a corresponding
        # type annotation.
        if (
            len([() for k in function.__annotations__ if k != 'return']) \
            != \
            len(inspect.getfullargspec(function).args)
        ):
            print(function.__code__.co_varnames)
            print(function.__annotations__)
            raise ValueError("function must have a type annotation for every argument")
        if 'return' not in function.__annotations__:
            raise ValueError("function must have an output type annotation")

        # Extract the type annotations and evaluate them for later processing.
        in_type = {k: eval_(a) for (k, a) in function.__annotations__.items() if k != 'return'}
        out_type = eval_(function.__annotations__['return'])

        # Ensure that every argument's type annotation is specified in a valid way.
        if not all(t is bit or isinstance(t, bits_type) for t in in_type.values()):
            raise TypeError("input type annotations must be specified using bit or bits")

        # Ensure that the output type annotation is specified in a valid way.
        if out_type is bit or isinstance(out_type, bits_type):
            out_type = (out_type,)
        elif isinstance(out_type, (tuple, list)):
            if not all(t is bit or isinstance(t, bits_type) for t in out_type):
                raise TypeError(
                    "output type annotation components must be specified using bit or bits"
                )
        else:
            raise TypeError(
                "output type must be specified using bit/bits, or a list/tuple thereof"
            )

    # Designate the circuit to be synthesized.
    bit.circuit(circuit())

    # Construct the input bit vector(s) and output wrapper based on the function's
    # type annotation or type information supplied via the parameters.
    in_values = {k: (input(0) if a is bit else inputs([0] * a)) for (k, a) in in_type.items()}
    out_values = output if out_type == (bit,) else outputs

    # Construct a signature to the circuit based on the type annotation of the
    # decorated function.
    signature_ = signature(
        [(1 if a is bit else a) for a in in_type.values()],
        [(1 if a is bit else a) for a in out_type]
    )

    # Synthesize the circuit by evaluating the function, add it to the function as an
    # attribute, and give the circuit its signature.
    try:
        out_values(function(**in_values))
        function_circuit = bit.circuit()
        function_circuit.signature = signature_
        if type_supplied_via_annotation:
            function.circuit = function_circuit
    except Exception as e:
        raise RuntimeError('automated circuit synthesis failed') from e

    # Return the original function if the function is being used as a decorator, or
    # the circuit if the function is being used programmatically (*i.e.*, with type
    # information supplied via its parameters).
    return function if type_supplied_via_annotation else function_circuit

if __name__ == "__main__":
    doctest.testmod() # pragma: no cover
